﻿using System;
using System.Collections.Generic;
using System.Dynamic;
using System.IO;
using System.Net;
using System.Text;

namespace com.Yavli.Resilience_Api
{
    class YavliResilience_Api
    {
        // explicitly declared properties
        /// <summary>
        /// Contains the locally stored Ad Server Code
        /// </summary>
        private string data = null;
        private dynamic data_decoded = null;

        private string website_id = null;
        private string api_key = null;
        private string api_url = "api.yavli.com/v2";
        private string cache_file = null;

        // implicitly declared properties in the constructor
        /// <summary>
        ///  The result obtained from the API stored as JSON string (as it was received without any modifications)
        /// </summary>
        private string api_result = null;
        private int? api_status = null;

        private bool? api_reachable = null;

        private string api_error_code = null;
        private string api_error_detail = null;

        /// <summary>
        /// Default constructor that sets up variables and detects available HTTPS methods.
        /// </summary>
        /// <param name="_cache_file"></param>
        public YavliResilience_Api(string _cache_file = "data.json")
        {
            cache_file = _cache_file;
            api_result = null;
            api_status = null;
            api_reachable = null;
            api_error_code = null;
            api_error_detail = null;
        }

        /// <summary>
        /// Sets the credentials of the client that uses this API to connect to Yavli.
        /// </summary>
        /// <param name="_website_id">The website Id.</param>
        /// <param name="_api_key">The API key.</param>
        public void set_credentials(string _website_id, string _api_key)
        {
            website_id = _website_id;
            api_key = _api_key;
        }

        /// <summary>
        /// Retrieves the latest version of the Yavli Ad Server Code from the Yavli API.
        /// </summary>
        /// <returns>True if the request was sucessful, otherwise False.</returns>
        public bool refresh_ad_server_code()
        {
            api_request("/get-ad-server");
            if (api_reachable == true && api_status == 200)
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// Fetch the locally stored Yavli Ad Server Code.
        /// </summary>
        /// <returns></returns>
        public string get_ad_server_code()
        {
            if (data_decoded == null)
            {
                load_data();
            }
            if (data_decoded != null)
            {
                return data_decoded.data.init_js;
            }
            throw new YavliException("No data found. You must run the API cron job before you can do this.");
        }

        /// <summary>
        /// Returns the Yavli API URL.
        /// </summary>
        /// <param name="path">The local path after the basic URL.</param>
        /// <returns>The full path to the API.</returns>
        private string get_api_url(string path)
        {
            return "http://" + api_url + path;
        }

        /// <summary>
        /// Loads the data from the local cache file and fills in the data property of this class.
        /// </summary>
        private void load_data()
        {
            if (String.IsNullOrEmpty(cache_file))
            {
                throw new YavliException("The property cache_file is null or empty. Provide a valid file name by calling the constructor of this function first.");
            }
            else if (System.IO.File.Exists(cache_file))
            {
                using (var sr = new System.IO.StreamReader(cache_file))
                {
                    data = sr.ReadToEnd();
                    data_decoded = Newtonsoft.Json.JsonConvert.DeserializeObject(data);
                }
            }
            else
            {
                throw new YavliException("The file:" + cache_file + " does not exist!");
            }
        }


        /// <summary>
        /// Stores the already downloaded Ad Server Code locally in the cache_file. This method can be replaced with one that stores the Ad Server code in a local database instead of a simple file.
        /// </summary>
        public void save_data()
        {
            if (api_reachable == true && api_status == 200)
            {
                try
                {
                    using (var sw = new System.IO.StreamWriter(cache_file))
                    {
                        sw.Write(api_result);
                    }
                }
                catch (Exception ex)
                {
                    throw new YavliException("Cache file is not writable: " + cache_file + " Inner Exception:" + ex.Message);
                }
                data = api_result;
            }
        }

        /// <summary>
        /// Returns the error message with all details.
        /// </summary>
        /// <returns></returns>
        public string get_error_msg()
        {
            StringBuilder msg = new StringBuilder();
            msg.Append("API Reachable: ");
            msg.Append(api_reachable == true ? "Yes" : "No");
            msg.Append(", API Status: '" + api_status + "', ");
            msg.Append("API Error Code: '" + api_error_code + "', ");
            msg.Append("API Error Detail: '" + api_error_detail + "'");
            return msg.ToString();
        }

        /// <summary>
        /// Returns the auth headers based on the provided WebsiteID and ApiKey
        /// </summary>
        private Dictionary<String, String> get_auth_headers()
        {
            if (website_id == null || api_key == null)
            {
                throw new YavliException("Credentials must be set in order to make API calls.");
            }
            Dictionary<String, String> result = new Dictionary<string, string>();
            result.Add("X-WebsiteID", website_id);
            result.Add("X-ApiKey", api_key);
            return result;
        }
        
        private static bool PropertyExists(ExpandoObject exo_result, string property_name)
        {
            return ((IDictionary<String, object>)exo_result).ContainsKey(property_name);
        }

        protected void api_request(string action)
        {
            string url = get_api_url(action);

            var hvars = get_auth_headers();
            string response = "";
            using (WebClient client = new WebClient())
            {
                try
                {
                    foreach (var kv in hvars)
                    {
                        client.Headers.Add(kv.Key, kv.Value);
                    }
                    byte[] ResponseBytes = client.DownloadData(url);
                    response = Encoding.UTF8.GetString(ResponseBytes);
                }
                catch (WebException wex)
                {
                    if (wex.Response != null)
                    {
                        var responseStream = wex.Response.GetResponseStream();
                        if (responseStream != null)
                        {
                            using (var reader = new StreamReader(responseStream))
                            {
                                response = reader.ReadToEnd();
                            }
                        }
                    }
                }
            }
            
            ExpandoObject exo_result = Newtonsoft.Json.JsonConvert.DeserializeObject<ExpandoObject>(response, new Newtonsoft.Json.Converters.ExpandoObjectConverter());
            dynamic dyn_result = Newtonsoft.Json.JsonConvert.DeserializeObject(response);
            if (dyn_result != null && exo_result != null)
            {
                api_reachable = true;
                if (PropertyExists(exo_result,"error"))
                {
                    api_status = dyn_result.error.status;
                    api_error_code = dyn_result.error.code;
                    api_error_detail = dyn_result.error.detail;
                }
                else
                {
                    api_status = 200;
                    api_result = response;
                }
            }
            else
            {
                api_reachable = false;
            }
        }
    }
}